@@ -29,36 +29,43 @@
 #ifndef __BOOT2_BOOT_H
 #define __BOOT2_BOOT_H
 
+#include "libsaio.h"
+
 /*
  * Keys used in system Default.table / Instance0.table
  */
 #define kGraphicsModeKey    "Graphics Mode"
+#define kTextModeKey        "Text Mode"
 #define kBootGraphicsKey    "Boot Graphics"
 #define kQuietBootKey       "Quiet Boot"
 #define kKernelFlagsKey     "Kernel Flags"
 #define kKernelNameKey      "Kernel"
-
-/*
- * Possible values for the bootdev argument passed to boot().
- */
-enum {
-	kBootDevHardDisk = 0,
-	kBootDevFloppyDisk,
-	kBootDevNetwork
-};
+#define kKernelCacheKey     "Kernel Cache"
 
 /*
  * A global set by boot() to record the device that the booter
  * was loaded from.
  */
-extern int gBootDev;
+extern int  gBIOSDev;
+extern BOOL gBootGraphics;
+extern long gBootMode;
+extern BOOL sysConfigValid;
+extern char bootBanner[];
+extern char bootPrompt[];
+
+// Boot Modes
+enum {
+    kBootModeNormal = 0,
+    kBootModeSafe,
+    kBootModeSecure
+};
 
 /*
  * graphics.c
  */
-extern void message(char * str, int centered);
-extern void setMode(int mode);
-extern int  currentMode();
+extern void printVBEInfo();
+extern void setVideoMode(int mode);
+extern int  getVideoMode();
 extern void spinActivityIndicator();
 extern void clearActivityIndicator();
 
@@ -66,5 +73,30 @@ extern void clearActivityIndicator();
  * drivers.c
  */
 extern long LoadDrivers(char * dirSpec);
+extern long DecodeKernel(void *binary, entry_t *rentry, char **raddr, int *rsize);
+
+/*
+ * options.c
+ */
+extern void getBootOptions();
+extern int  processBootOptions();
+
+/*
+ * lzss.c
+ */
+extern int decompress_lzss(u_int8_t *dst, u_int8_t *src, u_int32_t srclen);
+
+struct compressed_kernel_header {
+  u_int32_t signature;
+  u_int32_t compress_type;
+  u_int32_t adler32;
+  u_int32_t uncompressed_size;
+  u_int32_t compressed_size;
+  u_int32_t reserved[11];
+  char      platform_name[64];
+  char      root_path[256];
+  u_int8_t  data[0];
+};
+typedef struct compressed_kernel_header compressed_kernel_header;
 
 #endif /* !__BOOT2_BOOT_H */
