@@ -27,8 +27,9 @@
 #define __LIBSAIO_LIBSAIO_H
 
 #include "libsa.h"
-#include "saio.h"
+#include "memory.h"
+#include "io_inline.h"
 #include "saio_types.h"
 #include "saio_internal.h"
 
-#endif /* !__LIBSAIO_LIBSAIO_H */
\ No newline at end of file
+#endif /* !__LIBSAIO_LIBSAIO_H */
